# Version file automatically generated by hpc_connect

__static_version_tuple__: tuple[int, int, int, str] = (25, 8, 12, 'f9c4f4f')
__static_version__: str = "25.8.12+f9c4f4f"

def __getattr__(name):

    from hpc_connect.util.dynamic_version import version_components_from_git
    from hpc_connect.util.dynamic_version import GitRepoNotFoundError

    if name not in ("version", "__version__", "version_info", "__version_info__"):
        raise AttributeError(name)
    try:
        major, minor, micro, local = version_components_from_git(full=True)
        if name in ("version", "__version__"):
            return f"{major}.{minor}.{micro}+{local}"
        else:
            return (major, minor, micro, local)
    except GitRepoNotFoundError:
        if name in ("version", "__version__"):
            return __static_version__
        else:
            return __static_version_tuple__
